/***********************************************************************
#
#	PROCEDURE - Test Server
#           Send Messages to Echo Client and Receive Responses Using Poll
#
************************************************************************/
#define _GNU_SOURCE


#include <sys/types.h>
#include <poll.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/socket.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/utsname.h>
#include <netdb.h>
#include <unistd.h>



#define	NUM_FDS	1
#define	FNI_LISTEN_PORT		400
#define NUM_LISTEN  2	

extern int errno;
int server_fd = 0;
int client_fd = 0;
char errmsg[256];

main(argc,argv)
int argc;
char *argv[];
{
	int	nbytes, flags=0;
	void	(*sigfunc)(int), kpaliv(int), fni_term(int);
        struct  sigaction sigkeepalive;
 	struct	hostent *gethostbyname(), *ph;
        struct  sigaction sigterm;  
        char msg[256];
	struct	utsname name_info;
	struct	sockaddr_in raddr, *baddr;   
        sigterm.sa_flags = 0;
	struct	pollfd pfds[NUM_FDS];
        memset(errmsg, 0, 256);

	sigemptyset (&sigterm.sa_mask);
	sigemptyset (&sigterm.sa_mask);
        sigterm.sa_handler = fni_term;      
        sigterm.sa_flags = 0;

	if (sigaction(SIGTERM,&sigterm, NULL) == -1) {
                  sprintf(errmsg, "SIGACTION CALL FOR SIGTERM FAILED :  ", strerror(errno));
	          fprintf(stderr, errmsg);
		  fnip_exit(-1);
	}

        
	sigemptyset (&sigkeepalive.sa_mask);
        sigkeepalive.sa_handler = kpaliv;      
        sigkeepalive.sa_flags = 0;


	if (sigaction(SIGALRM,&sigkeepalive, NULL) == -1) {
                  sprintf(errmsg, "SIGACTION CALL FOR SIGALARM FAILED :  ", strerror(errno));
	          fprintf(stderr, errmsg);
		  fnip_exit(-1);
	}


    
        uname(&name_info);
 
        memset (&raddr, 0 ,sizeof(raddr));
	if ((ph = gethostbyname(name_info.nodename)) == NULL) {
                  sprintf(errmsg, "GETHOSTBYNAME CALL FAILED :  ", strerror(errno));
	          fprintf(stderr, errmsg);
		  fnip_exit(-1);
	}
	memcpy((caddr_t)&raddr.sin_addr,ph->h_addr,ph->h_length);
	raddr.sin_family =  AF_INET;
	raddr.sin_port = FNI_LISTEN_PORT;
	raddr.sin_port = htons(raddr.sin_port);
        baddr = &raddr;

        if ((server_fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP))== -1)
        {
                sprintf(errmsg, "Socket Error No: %s ", strerror(errno));
	        fprintf(stderr, errmsg);
		exit(-1);
        }   
 
	pfds[0].fd = server_fd;
        pfds[0].events = POLLIN;

        //Bind address 
        if (bind(server_fd, (struct sockaddr *) baddr, sizeof(raddr)) < 0) {
                sprintf(errmsg, "Bind Error No: %s ", strerror(errno));
	        fprintf(stderr, errmsg);
		exit(-1);
        }

       //Specify number of connections to accept.
        if(listen(server_fd, NUM_LISTEN) < 0) {
               sprintf(errmsg, "Listen Error No: %s ", strerror(errno));
	       fprintf(stderr, errmsg);
		exit(-1);
        }

        if((client_fd = accept(server_fd, NULL, NULL)) == -1) {
               sprintf(errmsg, "Accept Error No: %s ", strerror(errno));
	        fprintf(stderr, errmsg);
		exit(-1);
  		  }
	pfds[0].fd = client_fd;
	pfds[0].events = POLLIN;

        

	while (1) {
                sigset_t * sigmask;
                sigemptyset(sigmask);
                sigprocmask(SIG_SETMASK, NULL, sigmask);
                int ppret = 0;
              	ppret = ppoll(pfds,NUM_FDS,NULL, sigmask);
                if(ppret == -1) 
                {
		   if (errno == EINTR)
		     continue;
                   sprintf(errmsg, "Poll Error No: %s", strerror(errno));
		   fprintf(stderr, errmsg);
	  	   fnip_exit(-1);
		}

                int i = 0;
		for (i = 0; i < NUM_FDS; i++) {
                        sighold(SIGALRM);
  			switch(pfds[i].revents) {
			case 0:
				continue;
                        case POLLRDHUP:
                                sprintf(errmsg, "Hang Up Poll Event No: %s", strerror(pfds[i].revents));
		                fprintf(stderr, errmsg);
  				fnip_exit(-1);
 				break;
                                
			case POLLIN:
                               //Read Message and Echo
                               memset(&msg,0 ,256);
                               int readlen = 0;
		               if ((readlen = recv(client_fd, &msg, 256 ,0)) <= 0) {
                                 sprintf(errmsg, "Read Error No:%s", strerror(errno));
		                 fprintf(stderr,errmsg);
                                 return -1;
	                       }
                               int writelen = 0;
                               if((writelen = send(client_fd, &msg, readlen, 0)) <= 0) {
                                sprintf(errmsg, "Write Error No:%s", strerror(errno));
		                fprintf(stderr,errmsg);
                                return -1;
                               }
                               alarm(10);
                                break; 
			      default:
                                 sprintf(errmsg, "Unexpected Poll Event No:%i", pfds[i].revents);
		                 fprintf(stderr, errmsg);
 				 fnip_exit(-1);
                          } //End Switch
                          sigrelse(SIGALRM);
		}  //End -  For Loop
	}  //End of While Loop

}


fnip_exit(exit_status)

	int exit_status;
{
        close(server_fd);
 	exit(exit_status);
}

void fni_term(signum)
int signum; 
{

       fprintf(stderr, "Termination Signal Received");
       fnip_exit(0);
 }  

void kpaliv(signum)
  int signum;
{
        
	void	(*sigfunc)();
        struct  sigaction sigkeepalive;

	sigemptyset (&sigkeepalive.sa_mask);
        sigkeepalive.sa_handler = kpaliv;      
        sigkeepalive.sa_flags = 0;


	if (sigaction(SIGALRM,&sigkeepalive, NULL) == -1) {
                  sprintf(errmsg, "SIGACTION CALL FOR SIGALARM FAILED :  ", strerror(errno));
	          fprintf(stderr, errmsg);
        }
	alarm(10);
}
  
